@@ -2,20 +2,26 @@
 # -*- coding: utf-8 -*-
 #
 # ***** BEGIN LICENSE BLOCK *****
-# Copyright (C) 2012  Hayaki Saito <user@zuse.jp>
-#
-# This program is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# Copyright (C) 2012-2014, Hayaki Saito 
+# 
+# Permission is hereby granted, free of charge, to any person obtaining a 
+# copy of this software and associated documentation files (the "Software"), 
+# to deal in the Software without restriction, including without limitation 
+# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
+# and/or sell copies of the Software, and to permit persons to whom the 
+# Software is furnished to do so, subject to the following conditions: 
+# 
+# The above copyright notice and this permission notice shall be included in 
+# all copies or substantial portions of the Software. 
+# 
+# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
+# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
+# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL 
+# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
+# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
+# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
+# DEALINGS IN THE SOFTWARE. 
+# 
 # ***** END LICENSE BLOCK *****
 
 import sys
@@ -326,6 +332,7 @@ class DefaultPTY(PTY):
             os.close(self._master)
         except OSError, e:
             logging.exception(e)
+            logging.info("DefaultPTY.close: master=%d" % self._master)
 
     def restore_term(self):
         termios.tcsetattr(self._stdin_fileno,
@@ -430,97 +437,103 @@ class MockParseContext(ParseContext):
         output = StringIO()
         ParseContext.__init__(self, output)
 
+
 ###############################################################################
 #
-# Terminal
+# Process
 #
-class Terminal:
+class Process:
 
-    tty = None
+    _tty = None
+    _esc_timer = None
 
     def __init__(self, tty):
-        self.tty = tty
+        self._tty = tty
 
-    def start(self, termenc, inputhandler, outputhandler, inputparser, outputparser, listener):
+    def start(self, termenc,
+              inputhandler, outputhandler,
+              inputparser, outputparser,
+              inputscanner, outputscanner,
+              buffering=False,
+              stdout=sys.stdout):
 
-        inputcontext = ParseContext(output=self.tty,
+        inputcontext = ParseContext(output=self._tty,
                                     termenc=termenc,
-                                    scanner=DefaultScanner(),
+                                    scanner=inputscanner,
                                     handler=inputhandler,
-                                    buffering=False)
-        outputcontext = ParseContext(output=self.tty,
+                                    buffering=buffering)
+        outputcontext = ParseContext(output=stdout,
                                      termenc=termenc,
-                                     scanner=DefaultScanner(),
+                                     scanner=outputscanner,
                                      handler=outputhandler,
-                                     buffering=False)
+                                     buffering=buffering)
 
+        inputparser.init(inputcontext)
+        outputparser.init(outputcontext)
+
+        self._inputhandler = inputhandler
+        self._outputhandler = outputhandler
         self._inputparser = inputparser
         self._outputparser = outputparser
-        self._listener = listener
-        self._inputparser.init(inputcontext)
-        self._outputparser.init(outputcontext)
-        listener.handle_start(outputcontext)
+        self._inputcontext = inputcontext
         self._outputcontext = outputcontext
+        inputhandler.handle_start(outputcontext)
+        outputhandler.handle_start(outputcontext)
+
+    def getpid(self):
+        return self._tty.pid
+
+    def is_alive(self):
+        return self._tty is not None 
 
     def fileno(self):
-        return self.tty.fileno()
+        return self._tty.fileno()
+
+    def stdin_fileno(self):
+        return self._tty.stdin_fileno()
+
+    def read(self):
+        return self._tty.read()
+
+    def write(self, data):
+        self._tty.write(data)
 
     def close(self):
-        self.tty.close()
+        if self._tty is not None:
+            self._tty.close()
+            self._tty = None
 
     def end(self):
-        self._listener.handle_end(self._outputcontext)
+        self._inputhandler.handle_end(self._outputcontext)
+        self._outputhandler.handle_end(self._outputcontext)
+
+    def resize(self, row, col):
+        self._tty.resize(row, col)
+
+    def fitsize(self):
+        return self._tty.fitsize()
 
     def on_write(self, data):
         self._inputparser.parse(data)
 
-    def on_read(self, fd):
-        if fd == self.tty.fileno():
-            data = self.tty.read()
-            self._outputparser.parse(data)
+    def process_start(self):
+        self._inputhandler.handle_start(self._inputcontext)
+        self._outputhandler.handle_start(self._outputcontext)
+        self._inputhandler.handle_draw(self._outputcontext)
+        self._outputhandler.handle_draw(self._outputcontext)
+        #self._inputcontext.flush()
+        self._outputcontext.flush()
 
-###############################################################################
-#
-# Session
-#
-class Session:
+    def process_end(self):
+        self._inputhandler.handle_end(self._inputcontext)
+        self._outputhandler.handle_end(self._outputcontext)
 
-    def __init__(self, tty):
-
-        self._alive = True
-        self.terminal = Terminal(tty)
-        self.tty = tty
-        self._input_target = tty
-        main_master = self.tty.fileno()
-        stdin_fileno = self.tty.stdin_fileno()
-        self._rfds = [stdin_fileno, main_master]
-        self._wfds = []
-        self._xfds = [stdin_fileno, main_master]
-        self._resized = False
-        self._ttymap = {}
-
-    def add_subtty(self, term, lang, command, row, col,
-                   termenc, inputhandler,
-                   outputhandler, listener):
-
-        tty = DefaultPTY(term, lang, command, sys.stdin)
-        tty.resize(row, col)
-        process = Terminal(tty)
-
-        fd = tty.fileno()
-        self._rfds.append(fd)
-        self._xfds.append(fd)
-        self._ttymap[fd] = process
-
-        process.start(termenc,
-                      inputhandler,
-                      outputhandler,
-                      DefaultParser(),
-                      DefaultParser(),
-                      listener)
-
-        self._input_target = tty
-        return tty
+    def process_resize(self, row, col):
+        try:
+            self._inputhandler.handle_resize(self._inputcontext, row, col)
+            self._outputhandler.handle_resize(self._outputcontext, row, col)
+        finally:
+            self._resized = False
 
     def process_input(self, data):
         if not self._esc_timer is None:
@@ -546,18 +559,6 @@ class Session:
             self._esc_timer = threading.Timer(_ESC_TIMEOUT, dispatch_esc)
             self._esc_timer.start()
 
-    def process_start(self):
-        self._inputhandler.handle_start(self._inputcontext)
-        self._outputhandler.handle_start(self._outputcontext)
-        self._inputhandler.handle_draw(self._outputcontext)
-        self._outputhandler.handle_draw(self._outputcontext)
-        #self._inputcontext.flush()
-        self._outputcontext.flush()
-
-    def process_end(self):
-        self._inputhandler.handle_end(self._inputcontext)
-        self._outputhandler.handle_end(self._outputcontext)
-
     def process_output(self, data):
         self._outputparser.parse(data)
         self._inputhandler.handle_draw(self._outputcontext)
@@ -565,101 +566,137 @@ class Session:
         #self._inputcontext.flush()
         self._outputcontext.flush()
 
-    def process_resize(self, row, col):
-        try:
-            self._inputhandler.handle_resize(self._inputcontext, row, col)
-            self._outputhandler.handle_resize(self._outputcontext, row, col)
-        finally:
-            self._resized = False
+    def on_read(self, data):
+        self._outputparser.parse(data)
 
-    def focus_subprocess(self, tty):
-        self._input_target = tty
+    def drain(self):
+        self._inputparser.reset()
+        self._inputcontext.assign('')
 
-    def blur_subprocess(self):
-        self._input_target = self.tty
 
-    def destruct_subprocesses(self):
-        for fd in self._ttymap:
-            self.destruct_subprocess(fd)
+###############################################################################
+#
+# Session
+#
+class Session:
 
-    def destruct_subprocess(self, fd):
-        if fd in self._ttymap:
-            self._input_target = self.tty
-            if fd in self._rfds:
-                self._rfds.remove(fd)
-            if fd in self._xfds:
-                self._xfds.remove(fd)
-            process = self._ttymap[fd]
-            process.end()
-            process.close()
-            del self._ttymap[fd]
-            self.process_output("")
+    def __init__(self, tty):
+
+        self._alive = True
+        self._mainprocess = Process(tty)
+        self._input_target = self._mainprocess
+        stdin_fileno = self._mainprocess.stdin_fileno()
+        self._rfds = [stdin_fileno]
+        self._xfds = [stdin_fileno]
+        self._resized = False
+        self._process_map = {}
+
+    def add_subtty(self, term, lang,
+                   command, row, col,
+                   termenc,
+                   inputhandler=DefaultHandler(),
+                   outputhandler=DefaultHandler()):
+
+        tty = DefaultPTY(term, lang, command, sys.stdin)
+        tty.resize(row, col)
+        process = Process(tty)
+
+        self._init_process(process,
+                           termenc,
+                           inputhandler, outputhandler,
+                           DefaultParser(), DefaultParser(),
+                           DefaultScanner(), DefaultScanner(),
+                           buffering=False)
+        self.focus_process(process)
+        return process
+
+    def getactiveprocess(self):
+        return self._input_target
+
+    def process_is_active(self, process):
+        return self._input_target == process
+
+    def focus_process(self, process):
+        if process.is_alive():
+            logging.info("Switching focus: fileno=%d" % process.fileno())
+            self._input_target.drain()
+            self._input_target = process
+
+    def blur_process(self):
+        process = self._mainprocess
+        if process.is_alive():
+            logging.info("Switching focus: fileno=%d (main process)" % process.fileno())
+            self._input_target.drain()
+            self._input_target = process
+
+    def destruct_process(self, process):
+        fd = process.fileno()
+        if fd in self._rfds:
+            self._rfds.remove(fd)
+        if fd in self._xfds:
+            self._xfds.remove(fd)
+        process.end()
+        process.close()
+        del self._process_map[fd]
+
+        self.focus_process(self._mainprocess)
+        self._mainprocess.process_output("")
 
     def drive(self):
 
         def onresize(no, frame):
             ''' handle resize '''
             self._resized = True
+
         try:
             signal.signal(signal.SIGWINCH, onresize)
         except ValueError:
             pass
 
-        main_master = self.tty.fileno()
-        stdin_fileno = self.tty.stdin_fileno()
+        stdin_fileno = self._mainprocess.stdin_fileno()
         try:
             while self._alive:
-
                 try:
-                    rfds = self._rfds
-                    wfds = self._wfds
-                    xfds = self._xfds
-                    rfd, wfd, xfd = select.select(rfds, wfds, xfds)
+                    rfd, wfd, xfd = select.select(self._rfds, [], self._xfds, 0.6)
                     if xfd:
                         for fd in xfd:
-                            if fd == main_master:
-                                return
-                            elif fd == stdin_fileno:
-                                return
-                            else:
-                                if fd in self._ttymap:
-                                    self.destruct_subprocess(fd)
-                                    continue
-                                else:
-                                    return
+                            if fd in self._process_map:
+                                process = self._process_map[fd]
+                                self.destruct_process(process)
+                                continue
                     if self._resized:
                         self._resized = False
-                        row, col = self.tty.fitsize()
-                        self.process_resize(row, col)
-                    for fd in rfd:
-                        if fd == main_master:
-                            data = self.tty.read()
-                            self.process_output(data)
-                        elif fd == stdin_fileno:
-                            data = os.read(stdin_fileno, _BUFFER_SIZE)
-                            if self._input_target == self.tty:
-                                self.process_input(data)
-                            else:
-                                target_fd = self._input_target.fileno()
-                                process = self._ttymap[target_fd]
-                                process.on_write(data)
-                                self.process_input("")
-                        elif self._ttymap:
-                            ttymap = self._ttymap
-                            target_fd = self._input_target.fileno()
-                            if target_fd in ttymap:
-                                process = ttymap[target_fd]
-                                process.on_read(fd)
-                                self.process_output("")
+                        row, col = self._mainprocess.fitsize()
+                        self._mainprocess.process_resize(row, col)
+                    if rfd:
+                        for fd in rfd:
+                            if fd == stdin_fileno:
+                                data = os.read(stdin_fileno, _BUFFER_SIZE)
+                                if self._input_target.is_alive():
+                                    target_fd = self._input_target.fileno()
+                                    process = self._process_map[target_fd]
+                                    process.process_input(data)
+                                    self._mainprocess.process_input("")
+                            elif self._process_map:
+                                process_map = self._process_map
+                                if fd in process_map:
+                                    process = process_map[fd]
+                                    if self._input_target.is_alive():
+                                        if fd == self._input_target.fileno():
+                                            data = process.read()
+                                            process.on_read(data)
+                                            self._mainprocess.process_output("")
+                    else:
+                        pass
                 except select.error, e:
                     no, msg = e
-                    self._input_target = self.tty
                     if no == errno.EINTR:
                         # The call was interrupted by a signal
                         self._resized = True
                     elif no == errno.EBADF:
-                        for fd in self._ttymap:
-                            self.destruct_subprocess(fd)
+                        for fd in self._process_map:
+                            process = self._process_map[fd]
+                            self.destruct_process(process)
                     else:
                         raise e
                 except OSError, e:
@@ -679,15 +716,12 @@ class Session:
                 raise e
         finally:
             try:
-                self.process_end()
+                self._mainprocess.process_end()
             finally:
-                self.tty.close()
-                self._input_target = self.tty
-                for fd in self._ttymap:
-                    process = self._ttymap[fd]
+                for fd in self._process_map:
+                    process = self._process_map[fd]
                     process.end()
                     process.close()
-                self.ttymap = {}
 
     def start(self,
               termenc,
@@ -701,41 +735,53 @@ class Session:
               outputhandler=DefaultHandler(),
               buffering=False):
 
-        tty = self.tty
+        mainprocess = self._mainprocess
 
-        inputcontext = ParseContext(output=tty,
-                                    termenc=termenc,
-                                    scanner=inputscanner,
-                                    handler=inputhandler,
-                                    buffering=buffering)
-        outputcontext = ParseContext(output=stdout,
-                                     termenc=termenc,
-                                     scanner=outputscanner,
-                                     handler=outputhandler,
-                                     buffering=buffering)
-        inputparser.init(inputcontext)
-        outputparser.init(outputcontext)
+        self._init_process(mainprocess,
+                           termenc,
+                           inputhandler, outputhandler,
+                           inputparser, outputparser,
+                           inputscanner, outputscanner,
+                           buffering)
+        self.focus_process(mainprocess)
 
         self._resized = False
 
         def onclose(no, frame):
             pid, status = os.wait()
-            if pid == self.tty.pid:
+            if not mainprocess.is_alive():
                 self._alive = False
-            self._input_target = self.tty
+            elif pid == mainprocess.getpid():
+                self._alive = False
+            else:
+                self.focus_process(mainprocess)
 
         signal.signal(signal.SIGCHLD, onclose)
 
-        self._inputhandler = inputhandler
-        self._outputhandler = outputhandler
-        self._inputcontext = inputcontext
-        self._outputcontext = outputcontext
-        self._inputparser = inputparser
-        self._outputparser = outputparser
-
-        self._esc_timer = None
         self.drive()
 
+    def _init_process(self,
+                      process,
+                      termenc,
+                      inputhandler, outputhandler,
+                      inputparser, outputparser,
+                      inputscanner, outputscanner,
+                      buffering):
+
+        fd = process.fileno()
+        self._rfds.append(fd)
+        self._xfds.append(fd)
+        self._process_map[fd] = process
+
+        process.start(termenc,
+                      inputhandler,
+                      outputhandler,
+                      inputparser,
+                      outputparser,
+                      inputscanner,
+                      outputscanner,
+                      buffering)
+
 
 def _test():
     import doctest
